//
// 3dmf2wrl.c -- Converts a QuickDraw 3D Metafile to a VRML 97 file
//
// Copyright (C) 2000 by Bob Crispen <crispen@hiwaay.net> doing business as
//   Foxy Software
//
// LICENSE:
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef CONSOLE_VERSION
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "resource.h"
#include "convert.h"
#include "winfuncs.h"

HINSTANCE hInst;	// Instance handle
HWND hwndMain;		//nMain window handle

static char	szAppName[] = "3dmf2wrl";
static char     sz3dmfFileName[_MAX_PATH];
static char     szWrlFileName[_MAX_PATH];
static char     sz3dmfTitleName[_MAX_FNAME + _MAX_EXT];
static char     szWrlTitleName[_MAX_FNAME + _MAX_EXT];
static BOOL	bNeedSave;

#define UNTITLED "(no file)"

// Forward references
void DoCaption (HWND, char*);
void OkMessage (HWND, char*, char*);
short AskAboutSave (HWND, char*);

HWND hwndToolBar;
LRESULT CALLBACK MainWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);


/*------------------------------------------------------------------------
 Procedure:     MsgMenuSelect ID:1
 Purpose:       Shows in the status bar a descriptive explaation of
                the purpose of each menu item.The message
                WM_MENUSELECT is sent when the user starts browsing
                the menu for each menu item where the mouse passes.
 Input:         Standard windows.
 Output:        The string from the resources string table is shown
 Errors:        If the string is not found nothing will be shown.
------------------------------------------------------------------------*/
LRESULT MsgMenuSelect(HWND hwnd, UINT uMessage, WPARAM wparam, LPARAM lparam)
{
    static char szBuffer[256];
    UINT   nStringID = 0;
    UINT   fuFlags = GET_WM_MENUSELECT_FLAGS(wparam, lparam) & 0xffff;
    UINT   uCmd    = GET_WM_MENUSELECT_CMD(wparam, lparam);
    HMENU  hMenu   = GET_WM_MENUSELECT_HMENU(wparam, lparam);

    szBuffer[0] = 0;                            // First reset the buffer
    if (fuFlags == 0xffff && hMenu == NULL)     // Menu has been closed
        nStringID = 0;

    else if (fuFlags & MFT_SEPARATOR)           // Ignore separators
        nStringID = 0;

    else if (fuFlags & MF_POPUP)                // Popup menu
    {
        if (fuFlags & MF_SYSMENU)               // System menu
            nStringID = IDS_SYSMENU;
        else
            // Get string ID for popup menu from idPopup array.
            nStringID = 0;
    }  // for MF_POPUP
    else                                        // Must be a command item
        nStringID = uCmd;                       // String ID == Command ID

    // Load the string if we have an ID
    if (0 != nStringID)
        LoadString(hInst, nStringID, szBuffer, sizeof(szBuffer));
    // Finally... send the string to the status bar
    updateStatusBar(szBuffer);
    return 0;
}


// Initialize the window
static BOOL InitApplication(void)
{
	WNDCLASS wc;

	memset(&wc,0,sizeof(WNDCLASS));
	wc.style = CS_HREDRAW|CS_VREDRAW |CS_DBLCLKS;
	wc.lpfnWndProc = (WNDPROC)MainWndProc;
	wc.hInstance = hInst;
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
	wc.lpszClassName = "aWndClass";
	wc.lpszMenuName = MAKEINTRESOURCE(IDMAINMENU);
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	wc.hIcon = LoadIcon(hInst,MAKEINTRESOURCE(IDAPPLICON));
	if (!RegisterClass(&wc))
		return 0;
	return 1;
}

// Create the window
HWND CreateWndClassWnd(void)
{
	return CreateWindowEx(WS_EX_WINDOWEDGE|WS_EX_ACCEPTFILES,
		"aWndClass","3dmf2wrl",
		WS_MINIMIZEBOX|WS_VISIBLE|WS_CLIPSIBLINGS|WS_CLIPCHILDREN|WS_MAXIMIZEBOX|WS_CAPTION|WS_BORDER|WS_SYSMENU|WS_THICKFRAME,
		CW_USEDEFAULT,0,CW_USEDEFAULT,0,
		NULL,
		NULL,
		hInst,
		NULL);
}

// The "about" dialog
BOOL _stdcall AboutDlg(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {
		case WM_CLOSE:
			EndDialog(hwnd,0);
			return 1;
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDOK:
					EndDialog(hwnd,1);
					return 1;
			}
			break;
	}
	return 0;
}

// Handle the results of reading in the input file
void handle_result(HWND hwnd, int result) {
    switch (result) {
	case CNV_OK:
	    break;
	case CNV_CANT_OPEN:
	    OkMessage (hwnd, "Can't open %s!", sz3dmfTitleName);
	    sz3dmfFileName[0]  = '\0';
	    sz3dmfTitleName[0] = '\0';
	    break;
	case CNV_BAD_TRI_MESH_SPEC :
	    OkMessage (hwnd, "Bad triangle mesh spec in %s!", sz3dmfTitleName);
	    break;
	case CNV_CANT_ALLOC_NODE:
	    OkMessage (hwnd, "Error allocating memory in %s!", sz3dmfTitleName);
	    break;
	case CNV_BAD_TRIANGLE:
	    OkMessage (hwnd, "Bad coordinate index data in %s!", sz3dmfTitleName);
	    break;
	case CNV_BAD_POINTS:
	    OkMessage (hwnd, "Bad coordinate data in %s!", sz3dmfTitleName);
	    break;
	case CNV_EMPTY_FILE:
	    OkMessage (hwnd, "%s empty!", sz3dmfTitleName);
	    break;
	case CNV_BAD_COLOR:
	    OkMessage (hwnd, "Bad color data in %s!", sz3dmfTitleName);
	    break;
	case CNV_BAD_DIRECTION:
	    OkMessage (hwnd, "Bad forward direction data in %s!", sz3dmfTitleName);
	    break;
	case CNV_BAD_VIEWPOINT:
	    OkMessage (hwnd, "Bad viewpoint data in %s!", sz3dmfTitleName);
	    break;
	case CNV_BAD_LIGHT:
	    OkMessage (hwnd, "Bad light data in %s!", sz3dmfTitleName);
	    break;
    }
}


// Read in the file, and handle success
void readFile(HWND hwnd)
{
    int result;
    char* pExtension;
    static char szBuf[256];

    init_display();
    result = read_in(sz3dmfFileName);
    handle_result(hwnd, result);
    if (result == CNV_OK) {
	DoCaption (hwnd, sz3dmfTitleName);
	SendMessage(hwndToolBar, TB_ENABLEBUTTON,
	    (WPARAM)IDM_SAVEAS, (LPARAM)TRUE);
	sprintf(szBuf, "%s read OK", sz3dmfTitleName);
	updateStatusBar(szBuf);
	update_display(0, "");
	bNeedSave = TRUE;

	// Make a default VRML file name from the input file name
	memset(szWrlFileName, 0, _MAX_PATH);
	strcpy(szWrlFileName, sz3dmfFileName);
	pExtension = strrchr(szWrlFileName, '.');
	strcpy(pExtension, ".wrl");
	pExtension[4] = 0;
	GetFileTitle (szWrlFileName, szWrlTitleName,
	    sizeof (szWrlTitleName));
    }
}

// Main command handler
void MainWndProc_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
    static char szBuf[256];

	switch(id) {
		case IDM_OPEN:
		    if ((AskAboutSave (hwnd, szWrlTitleName)) == IDCANCEL)
			return;
		    if (FileOpenDlg (hwnd, sz3dmfFileName, sz3dmfTitleName)) {
			bNeedSave = FALSE;
			sprintf(szBuf, "Reading %s", sz3dmfTitleName);
			updateStatusBar(szBuf);
			UpdateWindow(hwnd);
			readFile(hwnd);
		    }
		    break;

		case IDM_SAVEAS :
		    if (FileSaveDlg (hwnd, szWrlFileName, szWrlTitleName)) {
			sprintf(szBuf, "Saving to %s", szWrlTitleName);
			updateStatusBar(szBuf);
			UpdateWindow(hwnd);
			if ((write_out(sz3dmfFileName)) == CNV_OK) {
			    bNeedSave = FALSE;
			    sprintf(szBuf, "VRML world saved in %s",
				szWrlTitleName);
			    updateStatusBar(szBuf);
			    update_display(0, "");
			    return;
			} else {
			    OkMessage (hwnd, "Could not write file %s",
				szWrlTitleName);
			}
		    }
		    break;

		case IDM_ABOUT:
		    DialogBox(hInst,MAKEINTRESOURCE(IDD_ABOUT),
			hwndMain,AboutDlg);
		    break;

		case IDM_EXIT:
		    PostMessage(hwnd,WM_CLOSE,0,0);
		    break;
	}
}

// Main procedure
LRESULT CALLBACK MainWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{

    LPTSTR szCmdLine;
    char* ptr;
    char* ptr2 = sz3dmfFileName;

    switch (msg) {

	case WM_CREATE:
		init_winfuncs(hwnd, hInst);

		// Get command line parameter if any
		memset(sz3dmfFileName, 0, sizeof(sz3dmfFileName));
		szCmdLine = GetCommandLine();
		ptr = szCmdLine;

		// Skip past the program name
		if (szCmdLine[0] == '"') {
		    ptr=strchr(&szCmdLine[1], '"');
		    if (ptr == NULL) {
			MessageBox (hwnd, "Bad command line!", szAppName,
			    MB_OK | MB_ICONEXCLAMATION);
			break;
		    }
		    ptr++;
		} else {
		    for (; ((!isspace(*ptr)) && (*ptr != '\0')); ptr++);
		}

		// Step ahead to the first non-whitespace character
		for (; isspace(*ptr) && (*ptr != '\0'); ptr++);
		if (*ptr == '"') {
		    // Quoted filename
		    for (++ptr; ((*ptr != '"') && (*ptr != '\0')); ptr++) {
			*ptr2++ = *ptr;
		    }
		} else {
		    // Naked filename
		    for (; (*ptr != '\0'); ptr++) {
			*ptr2++ = *ptr;
		    }
		}

		// If we ended up with a file name, try to open the file
		bNeedSave = FALSE;
		if (strlen (sz3dmfFileName) > 0) {
		    GetFileTitle (sz3dmfFileName, sz3dmfTitleName,
			sizeof (sz3dmfTitleName));
		    readFile(hwnd);
		}
		DoCaption (hwnd, sz3dmfTitleName);
		break;
	case WM_SIZE:
		{
		static RECT TBRect;
		int x, y, width, height;

		moveStatusBar(hwnd, msg, wParam, lParam);
		SendMessage(hwndToolBar,msg,wParam,lParam);
		GetWindowRect(hwndToolBar, &TBRect);
		x = 0;
		y = TBRect.bottom - TBRect.top;
		width = LOWORD(lParam) - x;
		height = HIWORD(lParam) - y - 5;
		moveNodeWindow(x, y, width, height);
		if (bNeedSave)
		    SendMessage(hwndToolBar, TB_ENABLEBUTTON,
			(WPARAM)IDM_SAVEAS, (LPARAM)TRUE);
		break;
		}
	case WM_MENUSELECT:
		return MsgMenuSelect(hwnd,msg,wParam,lParam);
	case WM_COMMAND:
		HANDLE_WM_COMMAND(hwnd,wParam,lParam,MainWndProc_OnCommand);
		break;
	case WM_CLOSE:
		if ((AskAboutSave (hwnd, szWrlTitleName)) == IDNO) {
		    DestroyWindow(hwnd);
		    clean_up();
		}
		return 0;
	case WM_DESTROY:
		clean_up();
		PostQuitMessage(0);
		break;
	case WM_NOTIFY:
		// Note: this may raise a warning about assigning a pointer
		// to an NMTREEVIEW to a pointer to an NMHDR.  Ignore it.
		// The lParam actually does contain a pointer to an NMTREEVIEW.
		HANDLE_WM_NOTIFY(hwnd, wParam, lParam, handle_node_msgs);
		break;

	default:
		return DefWindowProc(hwnd,msg,wParam,lParam);
	}
	return 0;
}

// Set up the toolbar
#define NUM_TOOLBAR_BUTTONS		2
HWND CreateAToolBar(HWND hwndParent)
{
        HWND hwndTB;
	TBADDBITMAP tbab;
	TBBUTTON tbb[NUM_TOOLBAR_BUTTONS];
	int iOpen, iSave;
	static char szBuf[256];

	// Ensure that the common control DLL is loaded.
	InitCommonControls();

	// Create a toolbar that the user can customize and that has a
	// tooltip associated with it.
	hwndTB = CreateWindowEx(0, TOOLBARCLASSNAME, (LPSTR) NULL,
	    WS_CHILD|WS_BORDER|CCS_ADJUSTABLE,
	    0, 0, 0, 0, hwndParent, (HMENU) ID_TOOLBAR, hInst, NULL);

	// Send the TB_BUTTONSTRUCTSIZE message, which is required for
	// backward compatibility.
	SendMessage(hwndTB, TB_BUTTONSTRUCTSIZE,
	    (WPARAM) sizeof(TBBUTTON), 0);

	// Add the bitmap containing button images to the toolbar.
	tbab.hInst = HINST_COMMCTRL;
	tbab.nID   = IDB_STD_SMALL_COLOR;
	SendMessage(hwndTB,TB_ADDBITMAP,NUM_TOOLBAR_BUTTONS,(LPARAM) &tbab);

	// clean memory before using it
	memset(tbb,0,sizeof tbb);

	// Add the strings
	memset(szBuf, 0, 255);
	LoadString(hInst, IDS_OPEN, szBuf, 255);
	iOpen = SendMessage(hwndTB, TB_ADDSTRING, 0, (LPARAM)(LPSTR)szBuf);
	memset(szBuf, 0, 255);
	LoadString(hInst, IDS_SAVEAS, szBuf, 255);
	iSave = SendMessage(hwndTB, TB_ADDSTRING, 0, (LPARAM)(LPSTR)szBuf);

	// Button "Open"
	tbb[0].iBitmap = STD_FILEOPEN;
	tbb[0].idCommand = IDM_OPEN;
	tbb[0].fsState = TBSTATE_ENABLED;
	tbb[0].fsStyle = TBSTYLE_BUTTON;
	tbb[0].iString = iOpen;

	// Button "Save"
	tbb[1].iBitmap = STD_FILESAVE;
	tbb[1].idCommand = IDM_SAVEAS;
	tbb[1].fsState = 0;
	tbb[1].fsStyle = TBSTYLE_BUTTON;
	tbb[1].iString = iSave;

	SendMessage(hwndTB,TB_ADDBUTTONS,NUM_TOOLBAR_BUTTONS,(LPARAM)&tbb);
	ShowWindow(hwndTB,SW_SHOW);
	return hwndTB;
}

// Do the window caption
void DoCaption (HWND hwnd, char *szTitleName)
{
	char szCaption[64 + _MAX_FNAME + _MAX_EXT];
	wsprintf (szCaption, "%s - %s", szAppName,
		szTitleName[0] ? szTitleName : UNTITLED);
	SetWindowText (hwnd, szCaption);
}

void OkMessage (HWND hwnd, char *szMessage, char *szTitleName)
{
	char szBuffer[64 + _MAX_FNAME + _MAX_EXT];
	wsprintf (szBuffer, szMessage, szTitleName[0] ? szTitleName : UNTITLED);
	MessageBox (hwnd, szBuffer, szAppName, MB_OK | MB_ICONEXCLAMATION);
}

short AskAboutSave (HWND hwnd, char *szTitleName)
{
	char szBuffer[64 + _MAX_FNAME + _MAX_EXT];
	int  iReturn;

	if (!bNeedSave)
	    return IDNO;

	wsprintf (szBuffer, "Save file in %s?",
		szTitleName[0] ? szTitleName : UNTITLED);
	iReturn = MessageBox (hwnd, szBuffer, szAppName,
		MB_YESNOCANCEL | MB_ICONQUESTION);

	if (iReturn == IDYES)
		if (!SendMessage (hwnd, WM_COMMAND, IDM_SAVEAS, 0L))
			iReturn = IDCANCEL;

	return iReturn;
}

/*<---------------------------------------------------------------------->*/
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, INT nCmdShow)
{
	MSG msg;
	HANDLE hAccelTable;

	hInst = hInstance;
	if (!InitApplication())
		return 0;
	hAccelTable = LoadAccelerators(hInst,MAKEINTRESOURCE(IDACCEL));
	if ((hwndMain = CreateWndClassWnd()) == (HWND)0)
	    return 0;
	hwndToolBar = CreateAToolBar(hwndMain);
	ShowWindow(hwndMain,SW_SHOW);
	UpdateWindow(hwndMain);
	while (GetMessage(&msg,NULL,0,0)) {
	    if (!TranslateAccelerator(msg.hwnd,hAccelTable,&msg)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	    }
	}
	return msg.wParam;
}

// Console version
#else
#include <stdio.h>
#include <string.h>
#include "convert.h"
#include "consolefuncs.h"

void usage()
{
    fprintf(stderr, "usage: 3dmf2wrl [3dmffile] [wrlfile]\n");
}

int main(int argc, char** argv)
{
    char mffilename[400];
    char wrlfilename[400];
    int argn = 1;
    int result;
    
    strcpy(mffilename, "stdin");
    strcpy(wrlfilename, "stdout");

    // -? -h[elp] = show syntax
    if (argc > 1) {
	if ((!strncmp(argv[1], "-?", 2)) || (!strncmp(argv[1], "-h", 2))) {
	    usage();
	    return 0;
	}
    }
    // Get input and output filenames if listed
    if (argn != argc) {
	strcpy(mffilename, argv[argn++]);
    }
    if (argn != argc) {
	strcpy(wrlfilename, argv[argn++]);
    }
    if (argn != argc) {
	usage();
	return 1;
    }
    // Read in the file
    result = read_in(mffilename);
    switch (result) {
	case CNV_OK:
	    break;
	case CNV_CANT_OPEN:
	    fprintf(stderr, "Can't open %s!\n", mffilename);
	    break;
	case CNV_BAD_TRI_MESH_SPEC :
	    fprintf(stderr, "Bad triangle mesh spec in %s!\n", mffilename);
	    break;
	case CNV_CANT_ALLOC_NODE:
	    fprintf(stderr, "Error allocating memory in %s!\n", mffilename);
	    break;
	case CNV_BAD_TRIANGLE:
	    fprintf(stderr, "Bad coordinate index data in %s!\n", mffilename);
	    break;
	case CNV_BAD_POINTS:
	    fprintf(stderr, "Bad coordinate data in %s!\n", mffilename);
	    break;
	case CNV_EMPTY_FILE:
	    fprintf(stderr, "%s empty!\n", mffilename);
	    break;
	case CNV_BAD_COLOR:
	    fprintf(stderr, "Bad color data in %s!\n", mffilename);
	    break;
	case CNV_BAD_DIRECTION:
	    fprintf(stderr, "Bad forward direction data in %s!\n", mffilename);
	    break;
	case CNV_BAD_VIEWPOINT:
	    fprintf(stderr, "Bad viewpoint data in %s!\n", mffilename);
	    break;
	case CNV_BAD_LIGHT:
	    fprintf(stderr, "Bad light data in %s!\n", mffilename);
	    break;
    }
    // Write out the file
    if ((write_out(wrlfilename)) != CNV_OK) {
	fprintf(stderr, "Error writing to %s\n", wrlfilename);
	return 1;
    }
    return 0;
}
#endif
