//
// vrml.h -- Definitions of VRML nodes and associated functions
//
// These functions do not assume anything about the platform, and should be
// usable on both the Windows version and the console version.
//
// Copyright (C) 2000 by Bob Crispen <crispen@hiwaay.net> doing business as
//   Foxy Software
//
// LICENSE:
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

// SF types
struct SFColor {
    float r;
    float g;
    float b;
};

struct SFImage {
    int width;
    int height;
    int numComponents;
    int* pixels;
};

struct SFVec2f {
    float s;
    float t;
};

struct SFVec3f {
   float x;
   float y;
   float z;
};

struct SFRotation {
    float x;
    float y;
    float z;
    float angle;
};

// Not a VRML type, but it's convenient
struct TriangleIndex {
    int first;
    int second;
    int third;
};

// MF types
struct MFColor {
    int numPoints;
    struct SFColor* v;
};

struct MFVec3f {
    int numPoints;
    struct SFVec3f* v;
};

struct MFVec2f {
    int numPoints;
    struct SFVec2f* v;
};

// Not a VRML type, but it's convenient
struct TriangleIndexArray {
    int numPoints;
    struct TriangleIndex* v;
};

// Nodes
struct Normal {
    struct MFVec3f vector;
};

struct Color {
    struct MFColor color;
};

struct Coordinate {
    struct MFVec3f point;
};

struct TextureCoordinate {
    struct MFVec2f point;
};

struct IndexedFaceSet {
    struct Color color;
    struct Coordinate coord;
    struct Normal normal;
    struct TextureCoordinate texCoord;
    int ccw;
    int* colorIndex;
    int colorPerVertex;
    int convex;
    struct TriangleIndexArray coordIndex;
    float creaseAngle;
    int* normalIndex;
    int normalPerVertex;
    int solid;
    int* texCoordIndex;
};

struct Material {
    float ambientIntensity;
    struct SFColor diffuseColor;
    struct SFColor emissiveColor;
    float shininess;
    struct SFColor specularColor;
    float transparency;
};

struct PixelTexture {
    struct SFImage image;
    int repeatS;
    int repeatT;
};

struct TextureTransform {
    struct SFVec2f center;
    float rotation;
    struct SFVec2f scale;
    struct SFVec2f translation;
};

// For this program, we just use PixelTextures
struct Appearance {
    struct Material material;
    struct PixelTexture texture;
    struct TextureTransform textureTransform;
};

// And the only kind of geometry node is an IFS
struct Shape {
    char* defName;
    struct Appearance appearance;
    struct IndexedFaceSet geometry;
};

// Viewpoint node
struct Viewpoint {
    char* defName;
    float fieldOfView;
    int   jump;
    struct SFRotation orientation;
    struct SFVec3f position;
    char* description;
};

// DirectionalLight node
struct DirectionalLight {
    char* defName;
    float ambientIntensity;
    struct SFColor color;
    struct SFVec3f direction;
    float intensity;
    int on;
};

// Functions
int numberOfShapes();
struct Shape* getShape(int);
struct Shape* newShape(char*, int, int);

int numberOfViewpoints();
int viewpointIsUnique(int);
struct Viewpoint* newViewpoint(char*);
struct Viewpoint* getViewpoint(int);

int numberOfLights();
int lightIsUnique(int);
struct DirectionalLight* newLight(char*);
struct DirectionalLight* getLight(int);

void newSceneGraph();
