//
// winfuncs.c -- Functions for the Windows version
//
// Copyright (C) 2000 by Bob Crispen <crispen@hiwaay.net> doing business as
//   Foxy Software
//
// LICENSE:
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <windows.h>
#include <commdlg.h>
#include <stdlib.h>
#include <stdio.h>
#include "resource.h"
#include "winfuncs.h"

static OPENFILENAME ofn;
// On programs that don't read in one format and export another,
// this second struct isn't necessary.  Don't forget to take out
// the code below that initializes it.
static OPENFILENAME ofn2;

// Handles to subwindows
HWND hwndProgress;
#define PROGRESS_WIDTH 98
HWND hwndStatusBar;
HWND hwndTV;
#define NUM_ICONS 3
#define ID_TREEVIEW 1
#define CX_ICON 16
#define CY_ICON 16
int g_nOpen, g_nClosed, g_nDocument;	// Indexes for bitmaps

// Forward references
HTREEITEM AddItemToTree(HWND, LPSTR, int);


// Initialize subwindows, file dialogs, etc.
void init_winfuncs(HWND hwnd, HINSTANCE hInst) {
    static char szFilter[] = "QuickDraw Files (*.3dmf)\0*.3dmf\0"  \
                             "All Files (*.*)\0*.*\0\0";
    static char szFilter2[] = "VRML Files (*.wrl)\0*.wrl\0"  \
                             "All Files (*.*)\0*.*\0\0";

    static RECT sbRect;		// Dimensions of status bar
    static int aWidths[2];	// Area widths for Progress bar
    RECT rcClient;		// Dimensions of client area 
    HIMAGELIST himl;		// Handle to image list for TreeView
    HBITMAP hbitmap;		// Handle to bitmap for TreeView


    // Ensure that the common control DLL is loaded. 
    InitCommonControls(); 

    // Create a status bar
    hwndStatusBar = CreateStatusWindow(
	WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | CCS_BOTTOM | SBARS_SIZEGRIP,
        "Ready", hwnd, IDM_STATUSBAR);
    GetWindowRect(hwndStatusBar, &sbRect);
    aWidths[0] = PROGRESS_WIDTH;
    aWidths[1] = -1;
    SendMessage(hwndStatusBar, SB_SETPARTS, 2, (LPARAM)aWidths);
    updateStatusBar("Ready");

    // Progress bar is a child of the status bar
    hwndProgress = CreateWindowEx (
	WS_EX_CLIENTEDGE,
	PROGRESS_CLASS,
	"",
	WS_CHILD | WS_VISIBLE,
	0,
	2,
	PROGRESS_WIDTH,
	sbRect.bottom - sbRect.top - 2,
	hwndStatusBar, 
	(HMENU)IDC_PROGRESS, 
	NULL,
	NULL);
    SendMessage(hwndProgress, PBM_SETRANGE, 0, MAKELPARAM(0, 100));

    // Get the dimensions of the parent window's client area, and create 
    // the tree view control. 
    GetClientRect(hwnd, &rcClient); 
    hwndTV = CreateWindowEx(0, WC_TREEVIEW, "Tree View", 
	WS_CLIPSIBLINGS | TVS_HASBUTTONS | TVS_LINESATROOT |
	WS_VISIBLE | WS_CHILD | WS_BORDER | TVS_HASLINES, 
	0, 0, rcClient.right, rcClient.bottom, 
	hwnd, (HMENU) ID_TREEVIEW, hInst, NULL); 

    // Create the image list. 
    himl = ImageList_Create(CX_ICON, CY_ICON, FALSE, NUM_ICONS, 0);

    // Add the open file, closed file, and document bitmaps. 
    hbitmap = LoadBitmap(hInst, MAKEINTRESOURCE(IDB_OPEN)); 
    g_nOpen = ImageList_Add(himl, hbitmap, 0); 
    DeleteObject(hbitmap); 

    hbitmap = LoadBitmap(hInst, MAKEINTRESOURCE(IDB_CLOSED)); 
    g_nClosed = ImageList_Add(himl, hbitmap, 0); 
    DeleteObject(hbitmap); 

    hbitmap = LoadBitmap(hInst, MAKEINTRESOURCE(IDB_NODE)); 
    g_nDocument = ImageList_Add(himl, hbitmap, 0); 
    DeleteObject(hbitmap); 

    // Associate the image list with the tree view control. 
    TreeView_SetImageList(hwndTV, himl, TVSIL_NORMAL); 

    // Set up the file I/O structs
    ofn.lStructSize       = sizeof (OPENFILENAME);
    ofn.hwndOwner         = hwnd;
    ofn.hInstance         = NULL;
    ofn.lpstrFilter       = szFilter;
    ofn.lpstrCustomFilter = NULL;
    ofn.nMaxCustFilter    = 0;
    ofn.nFilterIndex      = 0;
    ofn.lpstrFile         = NULL;          // Set in Open and Close functions
    ofn.nMaxFile          = _MAX_PATH;
    ofn.lpstrFileTitle    = NULL;          // Set in Open and Close functions
    ofn.nMaxFileTitle     = _MAX_FNAME + _MAX_EXT;
    ofn.lpstrInitialDir   = NULL;
    ofn.lpstrTitle        = NULL;
    ofn.Flags             = 0;             // Set in Open and Close functions
    ofn.nFileOffset       = 0;
    ofn.nFileExtension    = 0;
    ofn.lpstrDefExt       = "3dmf";
    ofn.lCustData         = 0L;
    ofn.lpfnHook          = NULL;
    ofn.lpTemplateName    = NULL;

    ofn2.lStructSize       = sizeof (OPENFILENAME);
    ofn2.hwndOwner         = hwnd;
    ofn2.hInstance         = NULL;
    ofn2.lpstrFilter       = szFilter2;
    ofn2.lpstrCustomFilter = NULL;
    ofn2.nMaxCustFilter    = 0;
    ofn2.nFilterIndex      = 0;
    ofn2.lpstrFile         = NULL;
    ofn2.nMaxFile          = _MAX_PATH;
    ofn2.lpstrFileTitle    = NULL;
    ofn2.nMaxFileTitle     = _MAX_FNAME + _MAX_EXT;
    ofn2.lpstrInitialDir   = NULL;
    ofn2.lpstrTitle        = NULL;
    ofn2.Flags             = 0;
    ofn2.nFileOffset       = 0;
    ofn2.nFileExtension    = 0;
    ofn2.lpstrDefExt       = "wrl";
    ofn2.lCustData         = 0L;
    ofn2.lpfnHook          = NULL;
    ofn2.lpTemplateName    = NULL;
}

// Move the status bar
void moveStatusBar (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    SendMessage(hwndStatusBar, msg, wParam, lParam);
}

// Update the message in the status bar
void updateStatusBar(char* msg)
{
    SendMessage(hwndStatusBar, SB_SETTEXT, 1, (LPARAM)msg);
}

// Move the node display window when required
void moveNodeWindow(int x, int y, int width, int height)
{
    MoveWindow(hwndTV, x, y, width, height, TRUE);
}

// Handle any messages coming to the node display window
void handle_node_msgs(HWND hwnd, WPARAM wParam, LPNMTREEVIEW ptv)
{
    TV_ITEM tvi;

    if (ptv->hdr.code == TVN_ITEMEXPANDING) {
	tvi.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE; 
	tvi.hItem = ptv->itemNew.hItem;
	tvi.state = ptv->itemNew.state;
	tvi.pszText = ptv->itemNew.pszText;
	tvi.lParam = ptv->itemNew.lParam;
	if (ptv->action == TVE_EXPAND) {
	    tvi.iImage = g_nOpen;
	    tvi.iSelectedImage = g_nOpen;
	} else {
	    tvi.iImage = g_nClosed;
	    tvi.iSelectedImage = g_nClosed;
	}
	TreeView_SetItem(hwndTV, &tvi); 
    }
}

// Show the file open dialog
BOOL FileOpenDlg (HWND hwnd, PSTR pstrFileName, PSTR pstrTitleName) {
    ofn.hwndOwner         = hwnd;
    ofn.lpstrFile         = pstrFileName;
    ofn.lpstrFileTitle    = pstrTitleName;
    ofn.Flags             = OFN_HIDEREADONLY | OFN_CREATEPROMPT;

    return GetOpenFileName (&ofn);
}

// Show the file save-as dialog
BOOL FileSaveDlg (HWND hwnd, PSTR pstrFileName, PSTR pstrTitleName) {
    ofn2.hwndOwner         = hwnd;
    ofn2.lpstrFile         = pstrFileName;
    ofn2.lpstrFileTitle    = pstrTitleName;
    ofn2.Flags             = OFN_OVERWRITEPROMPT;

    return GetSaveFileName (&ofn2);
}

// Initialize anything that needs to be initialized for the display
void init_display()
{
    TreeView_DeleteAllItems(hwndTV);
    AddItemToTree(hwndTV, "Top", 1);
}

// Update the display
void update_display(int pct, char* outputLine)
{

    // Show file progress
    SendMessage(hwndProgress, PBM_SETPOS, (WPARAM)pct, 0L);

    // Add message (if there is one) to display
    if (strlen(outputLine) > 0) {
	AddItemToTree(hwndTV, outputLine, 2);
    }
}

// Add an item to a TreeView
HTREEITEM AddItemToTree(HWND hwndTV, LPSTR lpszItem, int nLevel)
{
    TV_ITEM tvi; 
    TV_INSERTSTRUCT tvins; 
    static HTREEITEM hPrev = (HTREEITEM) TVI_FIRST; 
    static HTREEITEM hPrevRootItem = NULL; 
    static HTREEITEM hPrevLev2Item = NULL; 
    HTREEITEM hti; 

    tvi.mask = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM; 

    // Set the text of the item. 
    tvi.pszText = lpszItem; 
    tvi.cchTextMax = lstrlen(lpszItem); 

    // Assume the item is not a parent item, so give it a 
    // document image. 
    tvi.iImage = g_nDocument; 
    tvi.iSelectedImage = g_nDocument; 

    // Save the heading level in the item's application-defined 
    // data area. 
    tvi.lParam = (LPARAM) nLevel; 

    tvins.item = tvi; 
    tvins.hInsertAfter = hPrev; 

    // Set the parent item based on the specified level. 
    if (nLevel == 1) 
	tvins.hParent = TVI_ROOT; 
    else if (nLevel == 2) 
	tvins.hParent = hPrevRootItem; 
    else 
	tvins.hParent = hPrevLev2Item; 

    // Add the item to the tree view control. 
    hPrev = (HTREEITEM) SendMessage(hwndTV, TVM_INSERTITEM, 0, 
	(LPARAM) (LPTV_INSERTSTRUCT) &tvins); 

    // Save the handle to the item. 
    if (nLevel == 1) 
	hPrevRootItem = hPrev; 
    else if (nLevel == 2) 
	hPrevLev2Item = hPrev; 

    // The new item is a child item. Give the parent item a 
    // closed folder bitmap to indicate it now has child items. 
    if (nLevel > 1) { 
	hti = TreeView_GetParent(hwndTV, hPrev); 
	tvi.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE; 
	tvi.hItem = hti; 
	tvi.iImage = g_nClosed; 
	tvi.iSelectedImage = g_nClosed; 
	TreeView_SetItem(hwndTV, &tvi); 
    } 
    return hPrev;
}
